import re
import os
import click
import struct
import intelhex
import binascii
import subprocess
from fds_device_discovery import get_snr_of_connected_devices
from pynrfjprog import API
from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


HELP_TEXT_READ_TO_FILE = 'Read fds data to file (can be modified manually and re-written to device)'

fds_page_tag_common = '\xde\xad\xc0\xde'
fds_page_tag_common_LE32 = '\xde\xc0\xad\xde'  # Used when matching with data in flash.
fds_page_tag_swap_LE32 = fds_page_tag_common_LE32 + '\xFF\x01\x1E\xF1'
fds_page_tag_data_LE32 = fds_page_tag_common_LE32 + '\xFE\x01\x1E\xF1'
fds_page_tag_swap = fds_page_tag_common + '\xF1\x1E\x01\xFF'
fds_page_tag_data = fds_page_tag_common + '\xF1\x1E\x01\xFE'
fds_record_header_t = '<4HL'
fds_record_header_t_num_bytes = 4 * 2 + 4
fds_page_tag_size = 2 * 4
fds_page_tag_t = 'LL'

log_level = None


def _read_record_header(data_stream):
    r = dict()
    header = struct.unpack(fds_record_header_t, data_stream[:fds_record_header_t_num_bytes])
    r['record_key'] = header[0]
    r['len'] = header[1]
    r['file_id'] = header[2]
    r['crc'] = header[3]
    r['record_id'] = header[4]

    return r


def get_single_record_dict_from_binary_data(record_bin_str):
    try:
        # Record Key is first, and this is value indicates no more valid records in page
        if record_bin_str.startswith('\xff\xff'):
            return False
        r = _read_record_header(record_bin_str)
        if r['len'] > 0:
            # * 4 because len is word
            r['data'] = record_bin_str[fds_record_header_t_num_bytes:fds_record_header_t_num_bytes + r['len'] * 4]
            readable = binascii.hexlify(r['data'])
            readable = ' '.join([x for x in re.split(r'([A-Za-z0-9]{2})', readable) if len(x)])  # Split to 'xx xx xx'
            r['data_hex'] = readable
        return r
    except struct.error:  # Not enough data to unpack
        return False


def _get_num_bytes_in_record(record):
    return record['len'] * 4 + fds_record_header_t_num_bytes


def get_fds_dict_from_binary_fds_data(bin_str):
    fds_dict = dict()
    fds_dict['r'] = dict()
    fds_dict['num_pages'] = 0
    full_copy = bin_str

    found_fds_start_addr = 0
    while True:
        offset = bin_str.find(fds_page_tag_data_LE32)
        if not found_fds_start_addr:
            found_fds_start_addr = True
            fds_dict['start_addr'] = offset
        if offset == -1:  # Could not find FDS data page tag
            break
        fds_dict['num_pages'] += 1
        records_from_page = get_records_dict_from_single_fds_page_binary_data(bin_str[offset:])
        if records_from_page:
            offset += sum([_get_num_bytes_in_record(record) for record_id, record in records_from_page.iteritems()])
            # TODO Before updating, verify no key collision.
            fds_dict['r'].update(records_from_page)
        else:
            offset += fds_page_tag_size  # No records found, scan for page tag again.

        bin_str = bin_str[offset:]

    if fds_page_tag_swap_LE32 in full_copy and fds_page_tag_data_LE32 in full_copy:
        # Assume fds stores swap page first.
        swap_page_index = full_copy.find(fds_page_tag_swap_LE32)
        first_data_page_index = full_copy.find(fds_page_tag_data_LE32)
        if swap_page_index >= first_data_page_index:
            raise RuntimeError("Swap page located after first data page, stuff will break.")
        fds_dict['page_size'] = first_data_page_index - swap_page_index

    return fds_dict


def get_records_dict_from_single_fds_page_binary_data(fds_page_bin_str):
    r = dict()
    offset = 0
    while True:
        remaining_data = fds_page_bin_str[offset + fds_page_tag_size:]
        record = get_single_record_dict_from_binary_data(remaining_data)
        if not record:
            break
        if record['record_id'] != 0:  # 'record_id' == 0 indicates deleted item. Don't show these.
            if record['record_id'] in r.keys():
                raise RuntimeError("Duplicate record_id! %d" % record['record_id'])
            r[record['record_id']] = record
        num_bytes_read = _get_num_bytes_in_record(record)
        offset += num_bytes_read

    return r


def get_fds_dict_from_hex_file(hex_file):
    with open(hex_file) as f:
        i = intelhex.IntelHex(f)
        bin_str = i.tobinstr(start=0)  # To get correct 'start_addr' (which is based on index in this list), set start=0
        res = get_fds_dict_from_binary_fds_data(bin_str)
        return res


def get_fds_dict_from_yaml_file(yaml_filename, override_data_with_data_hex=True):
    with open(yaml_filename, 'r') as yaml_file:
        fds_dict = load(yaml_file, Loader=Loader)
        if override_data_with_data_hex:
            for rec_id, record in fds_dict['r'].iteritems():
                record['data'] = binascii.unhexlify(record['data_hex'].strip().replace(' ', ''))

        return fds_dict


def _crc16_data(data):
    hex_representation = binascii.hexlify(data)  # Need to perform hexlify as you cannot pass 'null' byte to shell.
    crc = int(subprocess.check_output('resources/crc16/crc16 %s' % hex_representation), base=16)
    return crc


def _crc16_record_dict(r):
    data_to_be_crced = struct.pack('<HHHL' + 'B' * len(r['data']),
                                   r['record_key'], r['len'], r['file_id'], r['record_id'],
                                   *[ord(x) for x in r['data']])
    return _crc16_data(data_to_be_crced)


def get_record_bin_data_from_dict(fds_dict):
    return get_record_bin_data(record_key=fds_dict['record_key'],
                               length=fds_dict['len'],
                               file_id=fds_dict['file_id'],
                               crc=fds_dict['crc'],
                               record_id=fds_dict['record_id'],
                               data=fds_dict['data'])


def get_record_bin_data(record_key, length, file_id, crc, record_id, data):
    return struct.pack(fds_record_header_t + 'B' * length * 4,
                       record_key, length, file_id, crc, record_id, *[ord(x) for x in data])


def create_fds_dict(record_key, length, file_id, crc, record_id, data):
    r = get_single_record_dict_from_binary_data(
        get_record_bin_data(record_key, length, file_id, crc, record_id, data))
    return r


def create_binary_fds_flash_data_from_fds_dict(fds_dict):
    # fds_data = fds_page_tag_data_LE32
    bytes_written_to_current_page = len(fds_page_tag_data)
    fds_pages = [fds_dict['num_pages']]
    current_page_idx = 0
    fds_pages[current_page_idx] = fds_page_tag_data_LE32
    for record_id, record in fds_dict['r'].iteritems():
        # TODO Check if data fits in page? (as only entry)
        bin_data = get_record_bin_data_from_dict(record)
        length_bytes = len(bin_data)

        # Check if we need to go to next page.
        if bytes_written_to_current_page + length_bytes > fds_dict['page_size']:
            # Fill rest of data.
            length_padding = fds_dict['page_size'] - bytes_written_to_current_page
            fds_pages[current_page_idx].join(['\xff' for _ in range(length_padding)])

            # Switch to next page, and write data tag.
            current_page_idx += 1
            bytes_written_to_current_page = len(fds_page_tag_data)
            fds_pages[current_page_idx] = fds_page_tag_data_LE32

        fds_pages[current_page_idx] += bin_data
        bytes_written_to_current_page += length_bytes

    # Create swap page
    swap_data = fds_page_tag_swap_LE32
    length_padding = fds_dict['page_size'] - len(fds_page_tag_swap)
    swap_data.join(['\xff' for _ in range(length_padding)])

    full_fds_data = swap_data + ''.join([x for x in fds_pages])
    return full_fds_data


def write_fds_dict_to_yaml_file(records_dict, out_file):
    output = dump(records_dict, Dumper=Dumper)

    with open(out_file, 'w') as out_file:
        out_file.write(output)


def _open_segger(segger_snr):
    if segger_snr is None:
        snrs = get_snr_of_connected_devices()
        if len(snrs) > 1:
            print('!Error, set active device by using "l" command first')
        else:
            segger_snr = snrs[0]
    if segger_snr is not None:
        with API.API(API.DeviceFamily.UNKNOWN) as api:
            api.connect_to_emu_with_snr(int(segger_snr))
            device_family = api.read_device_family()
        api = API.API(device_family)
        api.open()
        api.connect_to_emu_with_snr(int(segger_snr))
        return api


def get_fds_dict_from_binary_file(src):
    with open(src, 'rb') as source_binary_file:
        return get_fds_dict_from_binary_fds_data(source_binary_file.read())


def get_records_from_device(segger_snr, start_addr_for_search=0x50000):
        dev = _open_segger(segger_snr)
        bin_data = dev.read(start_addr_for_search, 0x80000)
        dev.close()
        records = get_fds_dict_from_binary_fds_data(''.join([chr(x) for x in bin_data]))
        records['start_addr'] += start_addr_for_search  # Need to add this as the existing value is based ons this.
        return records


def write_fds_dict_to_device(segger_snr, fds_dict, fix_crc=True):
    dev = _open_segger(segger_snr)

    if fix_crc:
        for record in fds_dict['r'].values():
            record['crc'] = _crc16_record_dict(record)

    data = create_binary_fds_flash_data_from_fds_dict(fds_dict)

    dev.write(fds_dict['start_addr'], [ord(x) for x in data], True)
    dev.close()


def write_fds_dict_to_bin_file(records_dict, dest):
    bin_data = create_binary_fds_flash_data_from_fds_dict(records_dict)
    ih = intelhex.IntelHex()
    ih.frombytes([ord(x) for x in bin_data])
    ih.tobinfile(dest)


def write_fds_dict_to_hex_file(records_dict, dest):
    bin_data = create_binary_fds_flash_data_from_fds_dict(records_dict)
    ih = intelhex.IntelHex()
    ih.frombytes([ord(x) for x in bin_data], offset=records_dict['start_addr'])
    ih.write_hex_file(dest)


def interactive():
    cmd = ''

    while cmd != 'x':
        # TODO
        cmd = raw_input(">: ")


def _resolve_snr(source):
    snr = None
    snrs = get_snr_of_connected_devices()
    if source.endswith('40') and len([x for x in snrs if x.startswith('682')]) == 1:
        snr = [x for x in snrs if x.startswith('682')][0]
    elif source.endswith('28') and len([x for x in snrs if x.startswith('681')]) == 1:
        snr = [x for x in snrs if x.startswith('681')][0]
    elif source.endswith('31') and len([x for x in snrs if x.startswith('680')]) == 1:
        snr = [x for x in snrs if x.startswith('680')][0]
    return snr

@click.group()
@click.option('-v', '--verbose',
              help='Show verbose information.',
              count=True)
def cli(verbose):
    pass


@cli.command()
def version():
    click.echo('I don\'t have a version yet, lazy developers.')


def is_pca(snr):
    return snr.lower() == 'pca10040' or snr.lower() == 'pca10028' or snr.lower() == 'pca10031'


@cli.command(short_help='Read fds data from device, hex file, bin file or yaml file')
@click.option('--source',
              help='Data to read. Segger serial number, path to .hex, .bin or .yaml file. If only one type of the board'
                   'is connected you can write "pca10040", "pca10028" or "pca10031" and the segger snr will be'
                   ' resolved automatically.',
              type=click.STRING)
@click.option('--dest',
              help='Where to store data. Segger serial number, path to .hex, .bin or .yaml file.'
                   'If not set, the result will be printed out.',
              type=click.STRING)
@click.option('--fds_start_addr',
              help='FDS start address. Default = 0',
              type=click.STRING)
def read(source, dest, fds_start_addr=None):  # TODO add support for overriding fds page data size and so on.
    if not source or source == dest:
        raise RuntimeError("Invalid args")
    fds_dict = None
    if re.match(r'\d{9}', source):
        # TODO Add address concept to all record get functions
        fds_dict = get_records_from_device(source)
    elif is_pca(source):
        snr = _resolve_snr(source.lower())
        if snr is None:
            raise RuntimeError("Unable to find correct device for %s" % source)
        else:
            fds_dict = get_records_from_device(snr)
    elif os.path.isfile(source):
        if source.endswith('.hex'):
            fds_dict = get_fds_dict_from_hex_file(source)
        elif source.endswith('.yaml'):
            fds_dict = get_fds_dict_from_yaml_file(source)
        elif source.endswith('.bin'):
            # NOTE! Address argument required in this case. No one has fds data at address 0
            if fds_start_addr is None:
                raise RuntimeError("Argument 'fds_start_addr' is required when reading from .bin file.")
            fds_dict = get_fds_dict_from_binary_file(source)
    else:
        raise RuntimeError('Invalid argument for "--in-data"')
    if fds_dict is None:
        click.echo('No records read! Nothing more to do here.')
        return 0
    if fds_start_addr is not None:
        fds_dict['start_addr'] = fds_start_addr

    if 'start_addr' not in fds_dict:
        raise RuntimeError("Could not find 'start_addr' key in records dict. This is required to be set. If reading"
                           "from binary file, the start_addr shall be provided as input argument.")

    click.echo('Found %d records.' % len(fds_dict['r']))
    if not dest:
        click.echo(fds_dict)
    elif re.match(r'\d{9}', dest) or dest == 'device':
        write_fds_dict_to_device(dest, fds_dict=fds_dict)
    elif is_pca(dest):
        write_fds_dict_to_device(_resolve_snr(dest), fds_dict=fds_dict)
    elif dest.endswith('.yaml'):
        write_fds_dict_to_yaml_file(fds_dict, dest)
        click.echo('Stored to ' + dest)
    elif dest.endswith('.hex'):
        write_fds_dict_to_hex_file(fds_dict, dest)
    elif dest.endswith('.bin'):
        write_fds_dict_to_bin_file(fds_dict, dest)
    else:
        click.echo('Unknown destination "%s"' % dest)


if __name__ == "__main__":
    DEBUGGING_WITH_PYCHARM = True

    if DEBUGGING_WITH_PYCHARM:
        import sys
        cli(sys.argv[1:])
    else:
        cli()
